/*

This code is copyright 2004-2012 by Michael Kohn
You may not use any of this program in your own programs
without permission from Michael Kohn <mike@mikekohn.net>

http://www.mikekohn.net/

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "avi.h"
#include "fileio.h"
#include "lzw.h"
#include "parse_bmp.h"

static char *extensions[] = { "gif", "bmp" };

static int write_avi_header(FILE *out, struct avi_header_t *avi_header)
{
int marker,t;

  write_chars_bin(out,"avih",4);
  marker=ftell(out);
  write_int32(out,0);

  write_int32(out,avi_header->time_delay);
  write_int32(out,avi_header->data_rate);
  write_int32(out,avi_header->reserved);
  write_int32(out,avi_header->flags);
  write_int32(out,avi_header->number_of_frames);
  write_int32(out,avi_header->initial_frames);
  write_int32(out,avi_header->data_streams);
  write_int32(out,avi_header->buffer_size);
  write_int32(out,avi_header->width);
  write_int32(out,avi_header->height);
  write_int32(out,avi_header->time_scale);
  write_int32(out,avi_header->playback_data_rate);
  write_int32(out,avi_header->starting_time);
  write_int32(out,avi_header->data_length);

  t=ftell(out);
  fseek(out,marker,SEEK_SET);
  write_int32(out,t-marker-4);
  fseek(out,t,SEEK_SET);

  return 0;
}

static int write_stream_header(FILE *out, struct stream_header_t *stream_header)
{
int marker,t;

  write_chars_bin(out,"strh",4);
  marker=ftell(out);
  write_int32(out,0);

  write_chars_bin(out,stream_header->data_type,4);
  write_chars_bin(out,stream_header->codec,4);
  write_int32(out,stream_header->flags);
  write_int32(out,stream_header->priority);
  write_int32(out,stream_header->initial_frames);
  write_int32(out,stream_header->time_scale);
  write_int32(out,stream_header->data_rate);
  write_int32(out,stream_header->start_time);
  write_int32(out,stream_header->data_length);
  write_int32(out,stream_header->buffer_size);
  write_int32(out,stream_header->quality);
  write_int32(out,stream_header->sample_size);
  write_int32(out,0);
  write_int32(out,0);

  t=ftell(out);
  fseek(out,marker,SEEK_SET);
  write_int32(out,t-marker-4);
  fseek(out,t,SEEK_SET);

  return 0;
}

static int write_stream_format(FILE *out, struct stream_format_t *stream_format)
{
int marker,t;

  write_chars_bin(out,"strf",4);
  marker=ftell(out);
  write_int32(out,0);

  write_int32(out,stream_format->header_size);
  write_int32(out,stream_format->width);
  write_int32(out,stream_format->height);
  write_int16(out,stream_format->num_planes);
  write_int16(out,stream_format->bits_per_pixel);
  write_int32(out,stream_format->compression_type);
  write_int32(out,stream_format->image_size);
  write_int32(out,stream_format->x_pels_per_meter);
  write_int32(out,stream_format->y_pels_per_meter);
  write_int32(out,stream_format->colors_used);
  write_int32(out,stream_format->colors_important);

  if (stream_format->colors_used!=0)
  {
    for (t=0; t<stream_format->colors_used; t++)
    {
      putc(stream_format->palette[t]&255,out);
      putc((stream_format->palette[t]>>8)&255,out);
      putc((stream_format->palette[t]>>16)&255,out);
      putc(0,out);
    }
  }

  t=ftell(out);
  fseek(out,marker,SEEK_SET);
  write_int32(out,t-marker-4);
  fseek(out,t,SEEK_SET);

  return 0;
}

static int write_junk_chunk(FILE *out)
{
int marker,t;
int r,l,p;
char *junk={ "JUNK IN THE CHUNK! " };

  write_chars_bin(out,"JUNK",4);
  marker=ftell(out);
  write_int32(out,0);

  r=4096-ftell(out);
  l=strlen(junk);
  p=0;

  for (t=0; t<r; t++)
  {
    putc(junk[p++],out);
    if (p>=l) p=0; 
  }

  t=ftell(out);
  fseek(out,marker,SEEK_SET);
  write_int32(out,t-marker-4);
  fseek(out,t,SEEK_SET);

  return 0;
}

static int write_avi_header_chunk(FILE *out,struct avi_header_t *avi_header,struct stream_header_t *stream_header,struct stream_format_t *stream_format)
{
int marker,t;
int sub_marker;

  write_chars_bin(out,"LIST",4);
  marker=ftell(out);
  write_int32(out,0);
  write_chars_bin(out,"hdrl",4);
  write_avi_header(out,avi_header);

  write_chars_bin(out,"LIST",4);
  sub_marker=ftell(out);
  write_int32(out,0);
  write_chars_bin(out,"strl",4);
  write_stream_header(out,stream_header);
  write_stream_format(out,stream_format);

  t=ftell(out);
  fseek(out,sub_marker,SEEK_SET);
  write_int32(out,t-sub_marker-4);
  fseek(out,t,SEEK_SET);

  write_junk_chunk(out);

  t=ftell(out);
  fseek(out,marker,SEEK_SET);
  write_int32(out,t-marker-4);
  fseek(out,t,SEEK_SET);

  return 0;
}

static int write_index(FILE *out, int count, int *offsets)
{
int marker,t;

  if (offsets==0) return -1;

  write_chars_bin(out,"idx1",4);
  marker=ftell(out);
  write_int32(out,0);

  for (t=0; t<count; t++)
  {
    write_chars(out,"rec ");
    write_int32(out,0x01);
    write_int32(out,offsets[t]);
    write_int32(out,offsets[t+1]-offsets[t]-12);

    write_chars(out,"00dc");
    write_int32(out,0x10);
    write_int32(out,offsets[t]+12);
    write_int32(out,offsets[t+1]-offsets[t]);
  }

  t=ftell(out);
  fseek(out,marker,SEEK_SET);
  write_int32(out,t-marker-4);
  fseek(out,t,SEEK_SET);

  return 0;
}

static int color_distance(int c1, int c2)
{
int d;

  /* note - this function is slooooow */

  d=abs((c1&255)-(c2&255))+abs(((c1>>8)&255)-((c2>>8)&255))+abs(((c1>>16)&255)-(
(c2>>16)&255));

  return d/3;
}

int get_color(int col, struct stream_format_t *stream_format)
{
int d,c;

  for (c=0; c<stream_format->palette_count; c++)
  {
    if (col==stream_format->palette[c]) return c;
  }

  if (stream_format->palette_count!=256)
  {
    stream_format->palette[stream_format->palette_count++]=col;
    return c;
  }

  /* if this part of the function is reached.. life will be slow */

  for (d=1; d<255; d++)
  {
    for (c=0; c<stream_format->palette_count; c++)
    {
      if (color_distance(col,stream_format->palette[c])<=d) return c;
    }
  }

  return 0;
}

static int build_stream_format_gif(char *dir, struct stream_format_t *stream_format)
{
FILE *in;
char filename[1024];
unsigned char temp_buffer[7];
struct gif_header_t gif_header;
int t;

  sprintf(filename,"%sframe%05d.gif",dir,0);
  printf("Trying to open %s\n", filename);
  in=fopen(filename,"rb");
  if (in==0) return -1;

  read_chars(in,(char *)temp_buffer,6);

  if (strcmp((char *)temp_buffer,"GIF89a")==0)
  { gif_header.version=89; }
    else
  if (strcmp((char *)temp_buffer,"GIF87a")==0)
  { gif_header.version=87; }
    else
  {
    printf("This is not a GIF.\n");
    return -1;
  }

  gif_header.width=read_int16(in);
  gif_header.height=read_int16(in);

  fclose(in);

  stream_format->width=gif_header.width;
  stream_format->height=gif_header.height;
  stream_format->colors_used=256;
  stream_format->colors_important=256;
  stream_format->palette=malloc(256*4);
  stream_format->palette_count=1;
  stream_format->bits_per_pixel=8;

  for (t=0; t<stream_format->colors_used; t++)
  {
    stream_format->palette[t]=0;
  }

  return 0;
}

static int build_stream_format_bmp(char *dir, struct stream_format_t *stream_format)
{
FILE *in;
char filename[1024];
int t;

  sprintf(filename,"%sframe%05d.bmp",dir,0);
  printf("Trying to open %s\n", filename);
  in=fopen(filename,"rb");
  if (in==0) return -1;

  if (getc(in)!='B') { fclose(in); return -1; }
  if (getc(in)!='M') { fclose(in); return -1; }

  fseek(in, 14+4, SEEK_SET);

  stream_format->width=read_int32(in);
  stream_format->height=read_int32(in);
  stream_format->colors_used=256;
  stream_format->colors_important=256;
  stream_format->palette=malloc(256*4);
  stream_format->palette_count=1;
  stream_format->bits_per_pixel=8;

  fclose(in);

  for (t=0; t<stream_format->colors_used; t++)
  {
    stream_format->palette[t]=0;
  }

  return 0;
}

static int count_frames(char *dir,int is_bmp)
{
FILE *in;
char filename[1024];
int t;

  printf("Scanning...\n");

  t=0;
  while(1)
  {
    sprintf(filename,"%sframe%05d.%s",dir,t,extensions[is_bmp]);
    in=fopen(filename,"rb");
    if (in==0) break;
    fclose(in);
    t++;
    if ((t%1000)==0) printf("%d frames so far...\n",t);
  }

  printf("Found %d frames\n",t);

  return t;
}

static int rle8_write_line(FILE *out,unsigned int *image,int width,struct stream_format_t *stream_format)
{
int count;
int r,t,k,out_count;

  out_count=0;
  t=0;

  while(t<width)
  {
    if (t==width-1)
    {
/*
printf("%d %d %d %d %d\n",image[635],image[636],image[637],
                          image[638],image[639]);
exit(0);
*/
      putc(1,out);
      putc(get_color(image[t],stream_format),out);
      t++; 
      break;
    }
      else
    if (image[t]==image[t+1])
    {
      count=1;
      k=t;
      while (image[t]==image[t+1])
      {
        count++;
        t++;
        if ((t+1)==width) break;
if (t>width) { printf("uh oh same %d %d %d\n",count,t,k); exit(0); }
        if (count==255) break;
      }

      putc(count,out);
      putc(get_color(image[t],stream_format),out);
      out_count=out_count+2;

// printf("%d %d %d\n",count,get_color(image[t],stream_format),t);

    }
      else
    {
      r=t;
      count=0;
      while (image[t]!=image[t+1])
      {
        count++;
        t++;
        if ((t+1)==width) break;
if (t>width) { printf("uh oh diff\n"); exit(0); }
        if (count==255) break;
      }

      if ((t+1)==width && count!=255) { count++; t++; }

      if (count==1)
      {
        putc(1,out);
        putc(get_color(image[r],stream_format),out);
        out_count+=2;
        continue;
      }
        else
      if (count==2)
      {
        putc(1,out);
        putc(get_color(image[r],stream_format),out);
        putc(1,out);
        putc(get_color(image[r+1],stream_format),out);
        out_count+=4;
        continue;
      }
        else
      {
        putc(0,out);
        // putc(t-r,out);
        putc(count,out);

if (t-r>255) { printf("more than 255\n"); exit(0); }
if (t-r!=count) { printf("not count\n"); exit(0); }

/*
        for (k=r; k<t; k++) 
        {
          putc(get_color(image[k],stream_format),out);
        }
*/
        t=r;
        for (k=0; k<count; k++) 
        {
          // putc(get_color(image[k],stream_format),out);
          putc(get_color(image[t++],stream_format),out);
        }
        if ((count%2)!=0) putc(0,out);

// printf(">> %d\n",count);

        out_count+=(2+count);
        continue;
      }
      t--;

    }
    t++;
  }

if (t!=640) printf("out_count=%d t=%d\n",out_count,t);
// exit(0);

  putc(0,out);
  putc(0,out);

/*
  if ((out_count%4)!=0)
  {
    r=(out_count%4);

    for (t=r; t<4; t++) putc(0,out);
  }
*/

  return 0;
}

static int write_movi(FILE *out,struct avi_header_t *avi_header,struct stream_header_t *stream_header,struct stream_format_t *stream_format,char *dir,int is_bmp)
{
FILE *in;
char filename[1024];
int width,height,depth,frame_count;
int marker,t,offsets_ptr,offsets_start;
int *offsets;
unsigned int *image;
int rec_marker,k;
int frame_size_marker;
int r;

/*
  width=320;
  height=240;
*/
  width=stream_format->width;
  height=stream_format->height;
  stream_format->bits_per_pixel=8;
  depth=stream_format->bits_per_pixel;
  if (depth==8)
  { stream_format->image_size=width*height; }
    else
  { stream_format->image_size=width*height/2; }
  frame_count=count_frames(dir,is_bmp);
  if (frame_count==0) return -1;

  offsets=malloc((frame_count+1)*4);

  avi_header->number_of_frames=frame_count;
  avi_header->width=width;
  avi_header->height=height;
  avi_header->buffer_size=stream_format->image_size;
  stream_header->buffer_size=stream_format->image_size;
  stream_header->data_length=frame_count;
  stream_format->bits_per_pixel=depth;

  image=malloc(width*height*4);

  write_chars_bin(out,"LIST",4);
  marker=ftell(out);
  write_int32(out,0);
  write_chars_bin(out,"movi",4);

  offsets_ptr=0;
  offsets_start=ftell(out)-4;

  for (t=0; t<frame_count; t++)
  {
    sprintf(filename,"%sframe%05d.%s",dir,t,extensions[is_bmp]);
    in=fopen(filename,"rb");
    if (in==0) break;

    offsets[offsets_ptr++]=ftell(out)-offsets_start;
    if (is_bmp==0)
    {
      lzw_decompress(in,image);
    }
      else
    {
      parse_bmp(in,image);
    }

    write_chars_bin(out,"LIST",4);
    rec_marker=ftell(out);
    write_int32(out,0);
    write_chars(out,"rec ");

    write_chars_bin(out,"00dc",4);
    /* write_int32(out,width*height); */
    frame_size_marker=ftell(out);
    write_int32(out,0);

    stream_format->compression_type=1;

    if (stream_format->compression_type==1)
    {
#ifdef BMPOUT
FILE *fp=fopen("out.bmp","wb+");
write_chars_bin(fp,"BM",2);
write_int32(fp,0);
write_int16(fp,0);
write_int16(fp,0);
write_int32(fp,0);   /* size */
write_int32(fp,40);            /* biSize */
write_int32(fp,width);
write_int32(fp,height);
write_int16(fp,1);
write_int16(fp,8);
write_int32(fp,1);             /* compression */
write_int32(fp,0);
write_int32(fp,0);             /* biXPelsperMetre */
write_int32(fp,0);             /* biYPelsperMetre */
write_int32(fp,256); /* biClrUsed */
write_int32(fp,256); /* biClrImportant */

for (t=0; t<256; t++)
// { write_int32(fp,stream_format->palette[t]); }
{ write_int32(fp,t*10); }

k=ftell(fp);
fseek(fp,10,SEEK_SET);
write_int32(fp,k);
fseek(fp,k,SEEK_SET);
#endif

      r=ftell(out);
      //for (k=0; k<height; k++)
      for (k=height-1; k>=0; k--)
      {
        rle8_write_line(out,image+(k*width),width,stream_format);
        // rle8_write_line(fp,image+(k*width),width,stream_format);
      }

      putc(0,out);
      putc(1,out);
      // putc(0,fp);
      // putc(1,fp);
/*
      r=ftell(out)-r;
      if ((r%4)!=0)
      {
        for (k=r; k<4; k++) putc(0,out);
      }
*/

#ifdef BMPOUT
k=ftell(fp);
fseek(fp,2,SEEK_SET);
write_int32(fp,k);
fseek(fp,k,SEEK_SET);

fclose(fp);

exit(0);
#endif
    }
      else
    if (depth==8)
    {
      for (k=0; k<width*height; k++)
      {
        putc(get_color(image[k],stream_format),out);
      }
    }
      else
    {
      for (k=0; k<width*height; k=k+2)
      {
        putc((get_color(image[k],stream_format)<<4)+(get_color(image[k+1],stream_format)),out);
      }
    }

    k=ftell(out);
    fseek(out,frame_size_marker,SEEK_SET);
    write_int32(out,(k-frame_size_marker)-4);
    fseek(out,k,SEEK_SET);

    k=ftell(out);
    fseek(out,rec_marker,SEEK_SET);
    write_int32(out,(k-rec_marker)-0);
    fseek(out,k,SEEK_SET);

    fclose(in);

    if ((t%100)==0) printf(".. %d frames complete %d%%\n",t,t*100/frame_count);
  }
  if ((t%100)==0) printf("%d frames complete 100%%\n",t);

  offsets[offsets_ptr++]=ftell(out)-offsets_start;

  t=ftell(out);
  fseek(out,marker,SEEK_SET);
  write_int32(out,t-marker-4);
  fseek(out,t,SEEK_SET);

  write_index(out,frame_count,offsets);

  free(image);
  free(offsets);

  return 0;
}

int write_avi(FILE *out, char *dir, int fps)
{
struct avi_header_t avi_header;
struct stream_header_t stream_header;
struct stream_format_t stream_format;
int is_bmp=0;
int t,marker;

  memset(&avi_header,0,sizeof(avi_header));
  memset(&stream_header,0,sizeof(stream_header));
  memset(&stream_format,0,sizeof(stream_format));

  write_chars_bin(out,"RIFF",4);
  write_int32(out,0);
  write_chars_bin(out,"AVI ",4);

  avi_header.time_delay=1000000/fps;
  avi_header.flags=0x10;
  avi_header.data_streams=1;
  strcpy(stream_header.data_type,"vids");
  strcpy(stream_header.codec,"DIB ");
  stream_header.time_scale=1;
  stream_header.data_rate=fps;
  stream_format.header_size=40;
  stream_format.num_planes=1;
  stream_format.compression_type=0;
  stream_format.palette=0;

  if (build_stream_format_gif(dir,&stream_format)==-1)
  {
    is_bmp=1;
    printf("Can't find gif's, trying bmp ...\n");
    if (build_stream_format_bmp(dir,&stream_format)==-1)
    {
      printf("Can't find bmp's either ... giving up.\n");
      return -1;
    }
  }

  printf("Found %s files with dimensions %dx%d\n", is_bmp==1?"BMP":"GIF", stream_format.width, stream_format.height);

  marker=ftell(out);
  write_avi_header_chunk(out,&avi_header,&stream_header,&stream_format);
  write_movi(out,&avi_header,&stream_header,&stream_format,dir,is_bmp);

  t=ftell(out);
  fseek(out,marker,SEEK_SET);
  write_avi_header_chunk(out,&avi_header,&stream_header,&stream_format);
  fseek(out,t,SEEK_SET);

  t=ftell(out);
  fseek(out,4,SEEK_SET);
  write_int32(out,t);
  fseek(out,t,SEEK_SET);

  if (stream_format.palette!=0) free(stream_format.palette);

  return 0;
}

int main(int argc, char *argv[])
{
FILE *out;
char dir[1024];
int fps;

  printf("\ngif2avi routines - Copyright 2005-2012 by Michael Kohn\n");
  printf("http://www.mikekohn.net/\n");
  printf("Version Date: 2012-Feb-09\n\n");

  if (argc!=4)
  {
    printf("This program will take a set of numbered GIF's (or BMP's) and\n");
    printf("decompress them in an avi movie file.\n\n");
    printf("Usage: gif2avi <outfile.avi> <directory with gif's> <frames per second>\n\n");
    exit(0);
  }

  fps=atoi(argv[3]);
  if (fps==0)
  {
    printf("Illegal value for frames per second\n");
    exit(1);
  }

  out=fopen(argv[1],"wb+");
  if (out==0)
  {
    printf("Could not open outfile %s\n",argv[1]);
    exit(1);
  }

  strcpy(dir,argv[2]);

  if (dir[strlen(dir)-1]!='/' && dir[strlen(dir)-1]!='\\')
  {
    strcat(dir,"/");
  }

  printf("Reading images from %s (%d fps)\n",dir,fps);

  write_avi(out,dir,fps);

  fclose(out);

  return 0;
}


